// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file CoarseningManager.h */

#ifndef AMDIS_COARSENINGMANAGER_H
#define AMDIS_COARSENINGMANAGER_H

// ============================================================================
// ===== includes =============================================================
// ============================================================================

#include "Global.h"
#include "Mesh.h"
#include "ProblemStatBase.h"

namespace AMDiS {

  // ============================================================================
  // ===== forward declarations =================================================
  // ============================================================================

  class Element;
  class Line;
  class Triangle;
  class Tetrahedron;
  class RCNeighbourList;
  class ElInfo;
  class TraverseStack;

  // ============================================================================
  // ===== class CoarseningManager ==============================================
  // ============================================================================

  /** \ingroup Adaption
   * \brief
   * Base class of CoarseningManager1d, CoarseningManager2d, CoarseningManager3d.
   * A CoarseningManager contains all functionality to perform coarsening
   * operations on the mesh. 
   */
  class CoarseningManager
  {
  public:
    /** \brief
     * constructs a CoarseningManager which belongs to aMesh 
     */
    CoarseningManager() 
      : mesh(NULL),  stack(NULL), globalMark(0), doMore(0) 
    {};

    /** \brief
     * destructor
     */
    virtual ~CoarseningManager() {};

    /** \brief
     * Returns the Mesh the CoarseningManager belongs to.
     */
    inline Mesh* getMesh() { return mesh; };

    //   /** \brief
    //    * Coarsens a single Element by decrement its mark and then coarsen
    //    * the mesh.
    //    */
    //   void coarsenElement(Element *element) {
    //     FUNCNAME("CoarseninManager::coarseElement");
    //     element->decrementMark();
    //     coarsenMesh();
    //   };

    /** \brief
     * Tries to coarsen every element of mesh at least mark times. First
     * all elements are marked for coarsening and then coarsenMesh will
     * be called. 
     */
    Flag globalCoarsen(Mesh *aMesh, int mark);
  
    /** \brief
     * Traversal routine for recursiv coarsening of a triangulation. It has
     * a default definition in CoarseningManager but it can be overriden
     * by sub classes (like in CoarseningManager1d), if another implementation
     * is needed. 
     */
    virtual Flag coarsenMesh(Mesh *aMesh);


  protected:
    /** \brief
     * If the mesh is coarsened by non recurisive traversal, this method 
     * spezifies
     * how one element of the mesh is coarsened. The method can be overriden
     * by sub classes, if used. 
     */
    virtual int coarsenFunction(ElInfo *) { return 0; };

    /** \brief
     *  Propagate coarsening information over the whole hierarchy
     *  by POSTORDER traversal of the hierarchy tree.       
     *  leaves:      'increment' coarsening mark,            
     *  inner nodes: set coarsening mark to                
     *               min(0,child[0].mark+1,child[1].mark+1)
     */
    void spreadCoarsenMark();

    /** \brief
     * Used while traversal in spreadCoarsenMark
     */
    static int spreadCoarsenMarkFunction(ElInfo* el_info);

    /** \brief
     *Used while traversal in cleanUpAfterCoarsen
     */
    static int cleanUpAfterCoarsenFunction(ElInfo* el_info);

    /** \brief
     * Sets the mark on all elements that have to be coarsend 
     */
    static int coarsenMarkFunction(ElInfo *el_info);

    /** \brief
     * Resets the element marks
     */
    void cleanUpAfterCoarsen();

  protected:
    /** \brief
     * the Mesh this CoarseningManager belongs to
     */
    Mesh   *mesh;

    /** \brief
     * Used for non recursive mesh traversal.
     */
    TraverseStack *stack;

    /** \brief
     * Used by globalCoarsen to remember the given mark value
     */
    int  globalMark;

    /** \brief
     * Spezifies whether the coarsening operation is still in progress
     */
    bool doMore;

    /** \brief
     * Used while mesh traversal to have a pointer to this CoarseningManager
     * from a static method
     */
    static CoarseningManager* traversePtr;

    /** \brief
     * Spezifies how many DOFVectors should restricted while coarsening
     */
    int callCoarseRestrict;

    friend class RCNeighbourList;
  };

}

#include "CoarseningManager1d.h"
#include "CoarseningManager2d.h"
#include "CoarseningManager3d.h"

#endif // AMDIS_COARSENINGMANAGER_H
